#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>

#include "hashtab.h"

#define MULTIPLIER 37  // multiplier for calculating the string hash function

struct entry *bkt_gen[BUCKSIZE] ;
struct entry *bkt_recv[BUCKSIZE] ;

struct entry *bkt_mgen[BUCKSIZE] ;
struct entry *bkt_mrecv[BUCKSIZE] ;

struct grp_entry *grp_bucket[MAX_GROUPS];
int    grp_size[MAX_GROUPS];

void init()
{
   int i;

   for ( i = 0; i < MAX_GROUPS; i++ ) {
      grp_size[i] = 0;
   }
}

struct entry *lookup ( logtype lt, const char *str, float time, bool insert )
{
   unsigned long h;
   int i, len;
   struct entry *p;

   struct entry **buckets;

   assert ( str ) ;

   len = strlen ( str ) ; 

   if ( len == 0 ) {
      return NULL ;
   }

   if ( lt == LOG_GEN ) {
      buckets = bkt_gen;
   } else if ( lt == LOG_RECV ) {
      buckets = bkt_recv;
   } else if ( lt == LOG_MGEN ) {
      buckets = bkt_mgen;
   } else if ( lt == LOG_MRECV ) {
      buckets = bkt_mrecv;
   } else {
      fprintf(stderr, "unknown log type\n");
      return NULL;
   }

   // hash function
   for ( h = 0, i = 0; i < len ; i++ ) {
      h = MULTIPLIER * h + str[i] ;
   }

   // trim value whithin bucket range
   h &= BUCKSIZE - 1 ;

   for ( p = buckets[h] ; p ; p = p->link ) {
      if ( len == p->len ) {
         if ( strcmp ( p->str, str ) == 0 ) {
	    return p ;
	 }
      }
   }

   // entry not found, insert it if asked (by 'insert')
   if ( insert == TRUE ) {
      p = (struct entry *) malloc ( sizeof ( struct entry ) ) ;
      p->len = len ;

      p->str = (char *) malloc ( len + 1 ) ;
      strcpy ( p->str, str ) ;
      p->str[len] = '\0' ;
      p->time = time ;
      p->flag = FALSE ;
      p->link = buckets[h] ;
      buckets[h] = p ;

      return p ;
   }

   return NULL ;
}

struct grp_entry *grp_lookup ( const int gid, const int nid, float time, bool insert ) 
{
   struct grp_entry *p;

   assert(gid < MAX_GROUPS);

   for ( p = grp_bucket[gid] ; p ; p = p->link ) {
      if ( nid == p->nid ) {
	 return p ;
      }
   }

   // entry not found, insert it if asked (by 'insert')
   if ( insert == TRUE ) {
      p = (struct grp_entry *) malloc ( sizeof ( struct grp_entry ) ) ;
      p->gid = gid ;

      p->nid = nid; 
      p->time = time ;
      p->link = grp_bucket[gid] ;
      grp_bucket[gid] = p ;

      grp_size[gid] += 1;

      return p ;
   }

   return NULL ;
}

void grp_print_all()
{
   int i;
   struct grp_entry *p ;

   for ( i = 0 ; i < MAX_GROUPS ; i++ ) {
      printf ( "group %d: ", i) ;

      for ( p = grp_bucket[i]; p ; p = p->link ) {
	 printf ( "%d ", p->nid ) ;
      }
      printf("\n");
   }
}

/*
 * print_all ( ) -- prints all the tokens stored in hash table
 */
void print_all ( logtype lt ) {
   struct entry *p ;
   int i, count = 0 ;

   struct entry **buckets;

   if ( lt == LOG_GEN ) {
      buckets = bkt_gen;
   } else if ( lt == LOG_RECV ) {
      buckets = bkt_recv;
   } else if ( lt == LOG_MGEN ) {
      buckets = bkt_mgen;
   } else if ( lt == LOG_MRECV ) {
      buckets = bkt_mrecv;
   } else {
      fprintf(stderr, "unknown log type\n");
      return;
   }

   for ( i = 0 ; i < BUCKSIZE ; i++ ) {
      for ( p = buckets[i]; p ; p = p->link ) {
	 printf ( "%s\t%f\n", p->str, p->time ) ;
         count++ ; 
      }
   }

   if ( count == 0 ) {
      printf ( "\nNo entries found.\n" ) ;
   } else {
      printf ( "\nNumber of entries found: %d.\n", count ) ;
   }
}

/*
 * print_stats() -- prints stats 
 */
void print_stats( bool verbose ) {
   struct entry *gp ;
   struct entry *rp ;

   int i;

   float avg_delay = 0.0; 
   float delay = 0.0;

   int total     = 0;
   int delivered = 0;

   for ( i = 0 ; i < BUCKSIZE ; i++ ) {
      for ( gp = bkt_gen[i]; gp ; gp = gp->link ) {
         total++;

         if ( (rp = lookup(LOG_RECV, gp->str, 0, FALSE)) == NULL ) { //generated message is not delivered
            if ( verbose ) {
               printf ( "m %s %f\n", gp->str, gp->time);
            }
         } else { // message is delivered
            delay = rp->time - gp->time;
            assert(delay >= 0.0);

            /* to avoid overflow, we turn avg = (n*avg + x)/(n+1) 
             * to equivalent form of: avg = (n/(n+1))*avg + x/(n+1) to avoid possible overflow of
             * the term n*avg + x, where n*avg is the total number by far, and x is the new value.
             */
            avg_delay = ((float)delivered/(delivered+1)) * avg_delay + delay/(delivered+1); 

            if ( verbose ) {
               printf("d %s %f %f %f\n", gp->str, gp->time, rp->time, delay); 
            }
            rp->flag = TRUE;

            delivered++;
         }
      }
   }

   if ( total == 0 ) { 
      printf("\nNo entries found.\n");
      return;
   }

   printf ( "Number of messages generated:    %d\n", total ) ;
   printf ( "Number of messages delivered:    %d\n", delivered ) ;
   printf ( "Average delay of delivered pkts: %f\n", avg_delay ) ;

   for ( i = 0 ; i < BUCKSIZE ; i++ ) {
      for ( rp = bkt_recv[i]; rp ; rp = rp->link ) {
          if ( rp->flag == FALSE ) {
               printf("ERROR: delivered, but not sent: %s (time %f)\n", rp->str, rp->time);
          }
      }
   }
}

/*
 * print_stats() -- prints stats 
 */
void print_mstats( bool verbose ) {
   struct entry *gp ;
   struct entry *rp ;

   // write full delivery delay (grp_dly) to the following temporary file
   FILE *ffdly = fopen("/tmp/scratchpad/fdly.txt", "w");

   int i, j;
   float gt;         // message generation time
   float grp_dly;  // delay till all the group members (e.g. the last group member) received the message
   float total_grp_dly = 0.0;

   int total_delivery[MAX_NODES];

   for ( i = 0; i < MAX_NODES; i++ ) {
      total_delivery[i] = 0;
   } 

   float avg_delay = 0.0; 
   float delay = 0.0;

   int total_gen  = 0;
   int delivered  = 0;
   int intended_delivered = 0;

   char tstr[128];

   int src;
   int seqno;
   int gid;
   int nid;

   int no_members;

   struct grp_entry *gen;
   int missed;
   int max_missed = 0;

   for ( i = 0 ; i < BUCKSIZE ; i++ ) {
      for ( gp = bkt_mgen[i]; gp ; gp = gp->link ) {
         total_gen++; // total number of messages generated

         gt = gp->time; // time this message generated
         grp_dly = 0.0;
         sscanf(gp->str, "%d:%d:%d", &src, &seqno, &gid);

         intended_delivered += grp_size[gid] - 1; // exclude source node itself from recepients

         no_members = 1; // source node is included
         for ( gen = grp_bucket[gid]; gen; gen = gen->link) {
            nid = gen->nid;

            if ( nid == src ) {
               continue;
            }

            memset(tstr, 0, sizeof(tstr));
            snprintf(tstr, sizeof(tstr), "%d:%d:%d", src, nid, seqno);

            if ( (rp = lookup(LOG_MRECV, tstr, 0, FALSE)) == NULL ) { //generated message is not delivered
               if ( verbose ) {
                  printf ( "m %s %f\n", gp->str, gp->time);
               }
            } else { // message is delivered
               delay = rp->time - gp->time;
               assert(delay >= 0.0);

               // we want to find out the delay until the last group member receives the message
               if ( delay > grp_dly ) {
                  grp_dly = delay;
               }

               no_members++;

               /* to avoid overflow, we turn avg = (n*avg + x)/(n+1) 
                * to equivalent form: avg = (n/(n+1))*avg + x/(n+1) to avoid possible overflow of
                * the term n*avg + x, where n*avg is the total number by far, and x is the new value.
                */
               avg_delay = ((float)delivered/(delivered+1)) * avg_delay + delay/(delivered+1); 

               if ( verbose ) {
                  printf("d %s %f %f %f\n", gp->str, gp->time, rp->time, delay); 
               }
               rp->flag = TRUE;

               delivered++;
            }
         }
         assert(no_members <= grp_size[gid]);
         missed = grp_size[gid] - no_members; // number of missed deliveries for this group message

         // if miss is zero, then it is a total delivery.
         if ( missed == 0 ) {
            total_grp_dly += grp_dly;
            if ( ffdly != NULL ) {
               fprintf(ffdly, "%f\n", grp_dly);
            }
         }
         total_delivery[missed] += 1;

         if ( missed > max_missed ) {
            max_missed = missed;
         }
      }
   }

   fclose(ffdly);

   if ( total_gen == 0 ) { 
      printf("\nNo entries found.\n");
      return;
   }

   printf ( "Number of messages generated:           %d\n", total_gen ) ;
   printf ( "Number of messages should be delivered: %d\n", intended_delivered ) ;
   printf ( "Number of messages delivered:           %d\n", delivered ) ;
   printf ( "Average delay of delivered pkts:        %f\n", avg_delay ) ;
   printf ( "Average delivery ratio (overall):       %f\n", ((float) delivered)/intended_delivered) ;
   printf ( "Average full delivery ratio:            %f\n", (total_delivery[0])/((float) total_gen));
   printf ( "Average full delivery delay:            %f\n", total_grp_dly/(total_delivery[0]) );
   
   float ratio;
   printf ( "\nDelivery distribution (no. packets delivered per multicast packet)\n" ) ;
   printf("\n");
   printf("no. missed  ratio\n");
   for ( i = 0; i <= max_missed; i++ ) {
      printf("%-3d %-5d", i, total_delivery[i]);
      no_members = total_delivery[i];
      ratio = ((float)no_members)/total_gen;
      //printf ( "%d\t", i) ;
      printf ( "  %.4f  ", ratio) ;
      for ( j = 0; j < (int) (100*ratio+0.5); j++ ) {
         printf("*");
      }
      printf("\n");
   }
   printf("\n");

   for ( i = 0 ; i < BUCKSIZE ; i++ ) {
      for ( rp = bkt_recv[i]; rp ; rp = rp->link ) {
          if ( rp->flag == FALSE ) {
               printf("ERROR: delivered, but not sent: %s (time %f)\n", rp->str, rp->time);
          }
      }
   }
}
